'''
This module contains functions to calculate evaluation metrics
'''

import numpy as np


def calc_mse(obs, sim):
    """Calculate the mean squared error.
    Args:
        obs: Array of the observed values
        sim: Array of the simulated values
    Returns:
        The MSE value for the simulation, compared to the observation.
    """
    # Validation check on the input arrays  
    if len(obs) != len(sim):
        raise ValueError("Arrays must have the same size.")
    # drop nan and negative temperature from the observation
    nan_index = obs>=0  
    obs = obs[nan_index]
    sim = sim[nan_index]
    
    # Calculate the rmse value
    mse_val = np.mean((obs-sim)**2)

    return mse_val

def calc_nse(obs, sim):
    """Calculate the Nash-Sutcliffe model efficiency coefficient.
    Args:
        obs: Array of the observed values
        sim: Array of the simulated values
    Returns:
        The NSE value for the simulation, compared to the observation.
    """
    
    if len(obs) != len(sim):
        raise ValueError("Arrays must have the same size.")
    
    # drop nan and negative temperature from the observation  
    nan_index = obs>=0  
    obs = obs[nan_index]
    sim = sim[nan_index]
    
    # denominator of the fraction term
    denominator = np.sum((obs-np.mean(obs))**2)
    
    # this would lead to a division by zero error and nse is defined as -inf
    if denominator == 0:
        raise ValueError("NSE denominator is 0")
    
    # numerator of the fraction term
    numerator = np.sum((sim-obs)**2)

    # calculate the NSE
    nse = 1 - numerator/denominator
    
    return nse

def calc_r2(obs, sim):
    
    if len(obs) != len(sim):
        raise ValueError("Arrays must have the same size.")
    
    # drop nan and negative temperature from the observation  
    nan_index = obs>=0  
    obs = obs[nan_index]
    sim = sim[nan_index]
    
    # denominator and numerator
    ss_res = np.sum((obs - sim)**2)
    ss_tot = np.sum((obs - np.mean(sim)) ** 2)
    
    # r2
    r2 = 1- (ss_res/ss_tot)
    return r2


# -*- coding: utf-8 -*-
# This file is part of RRMPG.
#
# RRMPG is free software with the aim to provide a playground for experiments
# with hydrological rainfall-runoff-models while achieving competitive
# performance results.
#
# You should have received a copy of the MIT License along with RRMPG. If not,
# see <https://opensource.org/licenses/MIT>

from .air2water import air2water


'''
The numba code for running the air2water model

'''

import numpy as np
from numba import njit

@njit()
def _a2w_8p(a1, a2, a3, a4, a5, a6, a7, a8,
            Ta, Tw, T_Ty, Th):
    # step wise calculate the temperature increment
    if Tw >= Th:
        delta = np.exp(-(Tw - Th)/a4)
        if delta == 0:
            delta = 1e-3
    else:
        delta = np.exp(-(Th - Tw)/a7) + np.exp(-Tw/a8)
    # delta has to be larger than 0
    k = (a1 + a2 * Ta - a3 * Tw + a5 * np.cos(2 * np.pi * (T_Ty - a6) ))/delta

    return k

@njit()
def _a2w_6p(a1, a2, a3, a4, a5, a6, 
            Ta, Tw, T_Ty, Th):
    # step wise calculate the temperature increment
    if Tw >= Th:
        delta = np.exp(-(Tw - Th)/a4)
        if delta == 0:
            delta = 1e-3
    else:
        delta = 1
    # delta has to be larger than 0
    k = (a1 + a2 * Ta - a3 * Tw + a5 * np.cos(2 * np.pi * (T_Ty - a6) ))/delta

    return k


@njit()
def run_air2water_8p(ta, 
                  t_ty,
                  th, 
                  tw_init, 
                  tw_ice,
                  params):
    """Implementation of the air2water model.
    
    
    The naming of the variables is kept as in the original publication [1].
    
    Args:
        tq: Numpy [t] array, which contains the airtemperature input.
        t_ty: Numpy [t] array, which contains the fraction of current day number to the total day number of the year
        th: Scalar for the deep water temperature
        tw_init: Scalar for the initial state of the water temperature.
        params: Numpy array of custom dtype, which contains the model parameter.
        
    Returns:
        tw: Numpy [t] array with the simulated streamflow.
            
    [1] Prediction of lake surface temperature using the air2water model: 
    guidelines, challenges, and future perspectives, 
    Advances in Oceanography and Limnology, 7:36-50, DOI: http://dx.doi.org/10.4081/aiol.2016.5791
        
    """    
    # Number of simulation timesteps
    num_timesteps = len(ta)
    
    # Unpack the model parameters
    a1 = params['a1']
    a2 = params['a2']
    a3 = params['a3']
    a4 = params['a4']
    a5 = params['a5']
    a6 = params['a6']
    a7 = params['a7']
    a8 = params['a8']

    # initialize empty arrays for lake surface water temperature
    tw = np.zeros(num_timesteps, np.float64)
    
    # set initial values
    tw[0] = tw_init
    
    # Start the model simulation
    # Use the forward RK45 explicit solution
    # time step dt = 1 day
    for t in range(1, num_timesteps):
        k1 = _a2w_8p(a1, a2, a3, a4, a5, a6, a7, a8,
                    ta[t-1], tw[t-1], t_ty[t-1], th)
        
        k2 = _a2w_8p(a1, a2, a3, a4, a5, a6, a7, a8,
                    (ta[t-1] + ta[t])/2, tw[t-1] + 0.5 * k1, (t_ty[t-1] + t_ty[t])/2, th)
        
        k3 = _a2w_8p(a1, a2, a3, a4, a5, a6, a7, a8,
                    (ta[t-1] + ta[t])/2, tw[t-1] + 0.5 * k2, (t_ty[t-1] + t_ty[t])/2, th)
        
        k4 = _a2w_8p(a1, a2, a3, a4, a5, a6, a7, a8,
                    ta[t], tw[t-1] + k3, t_ty[t], th)
        
        tw[t] = tw[t-1] + 1/6 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)

        # set the lower bound of the temperature when ice covered
        tw[tw < tw_ice] = tw_ice
    
    # return all but the artificial 0's step
    return tw

@njit()
def run_air2water_6p(ta, 
                  t_ty,
                  th, 
                  tw_init, 
                  tw_ice,
                  params):
    """Implementation of the air2water model.
    
    
    The naming of the variables is kept as in the original publication [1].
    
    Args:
        tq: Numpy [t] array, which contains the airtemperature input.
        t_ty: Numpy [t] array, which contains the fraction of current day number to the total day number of the year
        th: Scalar for the deep water temperature
        tw_init: Scalar for the initial state of the water temperature.
        params: Numpy array of custom dtype, which contains the model parameter.
        
    Returns:
        tw: Numpy [t] array with the simulated streamflow.
            
    [1] Prediction of lake surface temperature using the air2water model: 
    guidelines, challenges, and future perspectives, 
    Advances in Oceanography and Limnology, 7:36-50, DOI: http://dx.doi.org/10.4081/aiol.2016.5791
        
    """    
    # Number of simulation timesteps
    num_timesteps = len(ta)
    
    # Unpack the model parameters
    a1 = params['a1']
    a2 = params['a2']
    a3 = params['a3']
    a4 = params['a4']
    a5 = params['a5']
    a6 = params['a6']

    # initialize empty arrays for lake surface water temperature
    tw = np.zeros(num_timesteps, np.float64)
    
    # set initial values
    tw[0] = tw_init
    
    # Start the model simulation
    # Use the forward RK45 explicit solution
    # time step dt = 1 day
    for t in range(1, num_timesteps):
        k1 = _a2w_6p(a1, a2, a3, a4, a5, a6,
                    ta[t-1], tw[t-1], t_ty[t-1], th)
        
        k2 = _a2w_6p(a1, a2, a3, a4, a5, a6,
                    (ta[t-1] + ta[t])/2, tw[t-1] + 0.5 * k1, (t_ty[t-1] + t_ty[t])/2, th)
        
        k3 = _a2w_6p(a1, a2, a3, a4, a5, a6,
                    (ta[t-1] + ta[t])/2, tw[t-1] + 0.5 * k2, (t_ty[t-1] + t_ty[t])/2, th)
        
        k4 = _a2w_6p(a1, a2, a3, a4, a5, a6,
                    ta[t], tw[t-1] + k3, t_ty[t], th)
        
        tw[t] = tw[t-1] + 1/6 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)

        # set the lower bound of the temperature when ice covered
        tw[tw < tw_ice] = tw_ice
    
    # return all but the artificial 0's step
    return tw


'''
Util functions for air2water
'''

import numpy as np

# @njit
def get_param_bound(mean_depth, # precomputed depth
                        tw_range = (0, 30),
                        sradmax_range = (200, 450),
                        sradmin_range = (0, 200), 
                        albedo_range = (0.04, 0.2),
                        epsilon_a_range = (0.6, 0.9),
                        alpha_s_range = (3,15),
                        Delta_alpha_s_range = (0.1, 15),
                        ea_range = (5,15),
                        Delta_ea_range = (0.1, 10),
                        ):
    # Original script is written in MATLAB from https://github.com/marcotoffolon/air2water/    
    # This is adapted from its R version at https://github.com/aemon-j/air2wateR/blob/master/R/gen_param.R    
    n = 2
    empty_matrix = np.empty([n+1, n+1])
    Rad_a, Rad_b, ew_par = empty_matrix, empty_matrix, empty_matrix
    
    # Define constants
    rho = 1000 # km/m3
    cp = 4186 # J/kg K
    s_Boltzman = 5.67e-8 # W/m2 K4
    min_epsilon_a = 0.6
    bowen_ratio = 0.61
    
    # Get all possible daily max/min solar radiation, W/m2
    min_maxsrad, max_maxsrad = sradmax_range
    min_minsrad, max_minsrad = sradmin_range
    # delta_maxrad = max_maxsrad - min_maxsrad
    # delta_minrad = max_minsrad - min_minsrad
    maxrad = np.linspace(min_maxsrad, max_maxsrad, num = n+1)
    minrad = np.linspace(min_minsrad, max_minsrad, num = n+1)
    for i in range(n+1):
        for j in range(n+1):
            Rad_a[i,j] = (maxrad[i] - minrad[j]) * 0.5
            Rad_b[i,j] = (minrad[j] + maxrad[i]) * 0.5
    minRad_a, maxRad_a = np.min(Rad_a), np.max(Rad_a) # range of annual amplitude
    minRad_b, maxRad_b = np.min(Rad_b), np.max(Rad_b) # range of annual average
    
    # Plausible range for shortwave reflectivity (albedo) rs
    min_rs, max_rs = albedo_range
    rs = np.linspace(min_rs, max_rs, num = n+1)
    
    # Get all possible water temperature range, C
    min_Temperature_rif, max_Temperature_rif = tw_range
    Temperatura_rif = np.linspace(min_Temperature_rif, max_Temperature_rif, num = n+1)
    Kelvin0 = Temperatura_rif + 273.15
    min_Delta_T, max_Delta_T = tw_range
    Delta_T = np.linspace(min_Delta_T, max_Delta_T, num = n+1)
    
    # Plausible values for the emissivities of atmosphere (epsilon_a) and water (epsilon_w)
    min_epsilon_a, max_epsilon_a = epsilon_a_range
    epsilon_a = 0.97 * np.linspace(min_epsilon_a, max_epsilon_a, num = n+1) # (1-ra)*epsilon_a, where ra=0.03 (reflection of infrared radiation from water surface)
    epsilon_w = 0.97
    
    # Plausible values for the sensible/latent heat transfer functions (alpha_s/alpha_l, W/m2K)
    min_alpha_s, max_alpha_s = alpha_s_range
    alpha_s = np.linspace(min_alpha_s, max_alpha_s, num = n+1)
    alpha_l = alpha_s/bowen_ratio
    min_Delta_alpha_s, max_Delta_alpha_s = Delta_alpha_s_range
    Delta_alpha_s = np.linspace(min_Delta_alpha_s, max_Delta_alpha_s, num = n+1)
    Delta_alpha_l = Delta_alpha_s
    
    # Plausible values for the atmospheric water pressure (ea, mbar)
    min_ea, max_ea = ea_range
    ea = np.linspace(min_ea, max_ea, num = n+1)
    min_Delta_ea, max_Delta_ea = Delta_ea_range
    Delta_ea = np.linspace(min_Delta_ea, max_Delta_ea, num = n+1)
    
    # Plausible values for the water vapor saturation pressure at the temperature of water (ew, mbar)
    for i in range(n+1):
        for j in range(n+1):
            ew_par[i,j] = 6.112*np.exp(Temperatura_rif[i]*17.67/(Temperatura_rif[i]+243.5))
    ew = (np.max([0, np.min(ew_par)]), np.max(ew_par))
    
    # Reaction volume participating to the heat exchange with the atmosphere
    min_D = 1+mean_depth/20
    max_D = np.max([10, mean_depth])
    D = np.linspace(min_D, max_D, num = n+1)
    
    # Denominator
    den = rho * cp * D /86400 # conversino from seconds to days
    
    # parameter 1
    p1_par = np.empty([n+1, n+1, n+1, n+1, n+1, n+1, n+1, n+1], np.float64)
    for i in range(n+1):
        for j in range(n+1):
            for k in range(n+1):
                for l in range(n+1):
                    for m in range(n+1):
                        for o in range(n+1):
                            for p in range(n+1):
                                for q in range(n+1):
                                    temp = 6.112*np.exp(Temperatura_rif[m]*17.67/(Temperatura_rif[m]+243.5))*(1 - 17.67*243.5/(Temperatura_rif[m]+243.5)**2*Temperatura_rif[m] )
                                    p1_par[i,j,k,l,m,o,p,q] = ( (1-rs[p])*Rad_b[i,i] + s_Boltzman*Kelvin0[m]**3*(epsilon_a[j]-epsilon_w)*(273.15-3*Temperatura_rif[m])+alpha_l[k]*( ea[l] - temp ))/den[o]
                                    
    p1 = (np.min(p1_par), np.min([2, np.max(p1_par)]))
    
    # Parameter 2
    p2 = ((4*s_Boltzman*epsilon_a[0]*Kelvin0[0]**3 + alpha_s[0])/den[n], 
          (4*s_Boltzman*epsilon_a[n]*Kelvin0[n]**3 + alpha_s[n])/den[0] )
    
    # Parameter 3
    p3_par = np.empty([n+1, n+1, n+1, n+1, n+1, n+1, n+1, n+1], np.float64)
    for i in range(n+1):
        for j in range(n+1):
            for k in range(n+1):
                for l in range(n+1):
                        for o in range(n+1):
                            for p in range(n+1):
                                p3_par[i,j,k,l,o,p] = (4*s_Boltzman*epsilon_a[i]*Kelvin0[o]**3*( 1 - (epsilon_a[i]-epsilon_w)/epsilon_a[i])+alpha_s[p]+alpha_l[j]*6.112*np.exp(Temperatura_rif[k]*17.67/(Temperatura_rif[k]+243.5))*17.67*243.5/(Temperatura_rif[k]+243.5)**2)/den[l]
                                
                                
    p3 = (np.min(p3_par), np.max(p3_par))
    
    # Parameter 4
    p4 = (1, 100 * mean_depth**(-0.35))
    
    # Parameter 5
    p5 = ( ((1-rs[n])*minRad_a + alpha_l[0]*Delta_ea[0]+Delta_alpha_l[0]*(ea[0] - ew[1] + Delta_ea[0]) + Delta_alpha_s[0]*Delta_T[0] )/den[n],
           ((1-rs[0])*maxRad_a + alpha_l[n]*Delta_ea[n]+Delta_alpha_l[n]*(ea[n] - ew[0] + Delta_ea[n]) + Delta_alpha_s[n]*Delta_T[n] )/den[0]
          )
    
    # Parameter 6
    p6 = (0, 1)
    
    # Parameter 7
    p7 = (0, 150)
    
    # Parameter 8
    p8 = (0, 0.5)
    
    return p1, p2, p3, p4, p5, p6, p7, p8

def find_wider_range(bnd1, 
                     bnd2):
    # find low and high boundary
    bnd1_low, bnd1_high = bnd1
    bnd2_low, bnd2_high = bnd2
    # compare low bound
    low_bnd = np.min([bnd1_low, bnd2_low])
    high_bnd = np.max([bnd1_high, bnd2_high])
    
    # update the new bound
    new_bnd = (low_bnd, high_bnd)
    
    return new_bnd
    


from setuptools import setup, find_packages

long_description = "A python pacakge implementing an air2water model. Numba is used to speed up the computation. Parrallel computing is included for the calibration process. For detailed examples, check our Github repository: https://github.com/he134543/air2waterpy"

setup(
    name="air2waterpy",
    version="0.0.3",
    author="Xinchen He",
    author_email="xinchenhe@umass.edu",
    description="A python pacakge for running the air2water model",
    long_description= long_description,
    url="https://github.com/he134543/air2waterpy",
    packages=find_packages(),
    install_requires =[
        'numpy>=2.0.2',
        'pandas>=2.2.3',
        'pyswarms>=1.3.0',
        'numba>=0.60.0',
        'joblib>=1.4.2'
        ],
    license="MIT-License",
    classifiers=[
          'Programming Language :: Python :: 3.12',
          'License :: OSI Approved :: MIT License',
          'Topic :: Scientific/Engineering',
          'Intended Audience :: Science/Research'
        ]
    )


import numpy as np
import pandas as pd
from pyswarms.single.global_best import GlobalBestPSO
from joblib import Parallel, delayed
from functools import partial
from .air2water_model import run_air2water_6p, run_air2water_8p
from .metrics import calc_mse, calc_nse, calc_r2
from .gen_params import get_param_bound, find_wider_range


class air2water():
    """Interface to the the air2water model.
    This model implements the 6/8-parameter version of air2water model
    Original Publication:
        Piccolroaz S., M. Toffolon, and B. Majone (2013), 
        A simple lumped model to convert air temperature into surface water temperature in lakes, 
        Hydrol. Earth Syst. Sci., 17, 3323-3338, doi:10.5194/hess-17-3323-2013
    """
    def __init__(self,
                 params=None,
                 version="6p"):
        """Initialize an air2water model

        Args:
            params: (optional) Dictonary containing all model parameters as a seperate key/value pairs. Default to None
            version: (optional) String indicating which version is the air2water model. support "6p" and "8p". Default to "8p".
        """
        # load model version
        self._model_version = version
        if version == "6p":
            # List of model parameters
            self._param_list = ['a1', 'a2', 'a3', 'a4', 'a5', 'a6']

            # Initiate a default boundary based on physics meaning
            self._default_bounds = {'a1': (1e-4, 2),
                            'a2': (1e-4, 0.5),
                            'a3': (1e-4, 0.5),
                            'a4': (1, 50),
                            'a5': (1e-4, 1.2),
                            'a6': (0, 1)}
            # Custom numpy datatype needed for numba input
            self._dtype = np.dtype([('a1', np.float64),
                            ('a2', np.float64),
                            ('a3', np.float64),
                            ('a4', np.float64),
                            ('a5', np.float64),
                            ('a6', np.float64)])
        
        elif version == "8p":
            # List of model parameters
            self._param_list = ['a1', 'a2', 'a3', 'a4', 'a5', 'a6', 'a7', 'a8']

            # Initiate a default boundary based on physics meaning
            self._default_bounds = {'a1': (1e-4, 2),
                            'a2': (1e-4, 0.5),
                            'a3': (1e-4, 0.5),
                            'a4': (1, 50),
                            'a5': (1e-4, 1.2),
                            'a6': (0, 1),
                            'a7': (0, 150),
                            'a8': (0, 0.5)}
            # Custom numpy datatype needed for numba input
            self._dtype = np.dtype([('a1', np.float64),
                            ('a2', np.float64),
                            ('a3', np.float64),
                            ('a4', np.float64),
                            ('a5', np.float64),
                            ('a6', np.float64),
                            ('a7', np.float64),
                            ('a8', np.float64)])
        else:
            raise ValueError("Select the correct model version. Choose either '6p' or '8p'.")
        
        # Randomly generate a parameter if no params was passed
        if params == None:
            num = 1
            params = np.zeros(num, dtype=self._dtype)
            # sample one value for each parameter
            for param in self._param_list:
                values = np.random.uniform(low=self._default_bounds[param][0],
                                high=self._default_bounds[param][1],
                                size=num)
                params[param] = values
            self.params = params
        elif isinstance(params, dict):
            self.load_param(params)
        else:
            raise ValueError("Set params as None or use python dictionary")
        
    def load_params(self, params):
        """A function to load parameters into the model

        Args:
            params (Dict): a python dictionary contain model parameters
        """
        num = 1
        params_dict = params.copy()
        params = np.zeros(num, dtype=self._dtype)
        for key, value in params_dict.items():
            params[key] = value
        # load parameters
        self.params = params
            
        
    def update_param_bnds(self, 
                        mean_depth_range,
                        tw_range = (0, 30),
                        sradmax_range = (200, 450),
                        sradmin_range = (0, 200), 
                        albedo_range = (0.04, 0.2),
                        epsilon_a_range = (0.6, 0.9),
                        alpha_s_range = (3,15),
                        Delta_alpha_s_range = (0.1, 15),
                        ea_range = (5,15),
                        Delta_ea_range = (0.1, 10),
                        ):
        """A function to estimate the potential parameter range based on the lake characteristics. 
        This function is adapted from the matlab (https://github.com/spiccolroaz/air2water/) and also the R script (https://github.com/aemon-j/air2wateR)
        Instead of directly give a mean lake depth, this function allows providing a range of potential lake mean depth.

        Args:
            mean_depth_range (_tuple_): _Possible range of the estimated depth_
            tw_range (tuple, optional): _Range of all possible temperature_. Defaults to (0, 30).
            sradmax_range (tuple, optional): _Range of daily maximum shortwave solar radiation_. Defaults to (200, 450).
            sradmin_range (tuple, optional): _Range of daily minimum shortwave solar radiation_. Defaults to (0, 200).
            albedo_range (tuple, optional): _Range of daily minimum shortwave solar radiation_. Defaults to (0.04, 0.2).
            epsilon_a_range (tuple, optional): _Range of epsilon a_. Defaults to (0.6, 0.9).
            alpha_s_range (tuple, optional): _Range of albedo_. Defaults to (3,15).
            Delta_alpha_s_range (tuple, optional): _description_. Defaults to (0.1, 15).
            ea_range (tuple, optional): _description_. Defaults to (5,15).
            Delta_ea_range (tuple, optional): _description_. Defaults to (0.1, 10).
        """
        mean_depth_low, mean_depth_high = mean_depth_range
        # Calculate the parameter bnds based on different depth estimate
        bnds_deep = get_param_bound(mean_depth=mean_depth_high, 
                                    tw_range=tw_range, 
                                    sradmax_range=sradmax_range, 
                                    sradmin_range=sradmin_range, 
                                    albedo_range=albedo_range,
                                    epsilon_a_range=epsilon_a_range,
                                    alpha_s_range=alpha_s_range,
                                    Delta_alpha_s_range=Delta_alpha_s_range,
                                    ea_range=ea_range,
                                    Delta_ea_range=Delta_ea_range
                                    )
        bnds_shallow = get_param_bound(mean_depth=mean_depth_low,
                                    tw_range=tw_range, 
                                    sradmax_range=sradmax_range, 
                                    sradmin_range=sradmin_range, 
                                    albedo_range=albedo_range,
                                    epsilon_a_range=epsilon_a_range,
                                    alpha_s_range=alpha_s_range,
                                    Delta_alpha_s_range=Delta_alpha_s_range,
                                    ea_range=ea_range,
                                    Delta_ea_range=Delta_ea_range)
        
        # Choose the widest range of each parameter
        if self._model_version == "8p":
            for i in range(8):
                self._default_bounds[self._param_list[i]] = find_wider_range(bnds_deep[i], bnds_shallow[i])
        elif self._model_version == "6p":
            for i in range(6):
                self._default_bounds[self._param_list[i]] = find_wider_range(bnds_deep[i], bnds_shallow[i])
        else:
            raise ValueError("Select model version '6p' or '8p'")

    def simulate(self,
                 ta,
                 period,
                 th = 4.0,
                 tw_init = 1.0,
                 tw_ice = 0.0, 
                 ):
        """Simulate the lake surface water temperature for the passed air temperature. This function is adapted from the RRMPG: https://github.com/kratzert/RRMPG.git
        
        Args:
            ta (pandas.Series): air temperature data for each timestep. Preferably numpy array, could be a list or pandas series.
            period (pandas.Series): simulation period, use pd.Timestamp to indicate time. Preferably pandas series.
            th: deep water temperature. Const
            tw_init: (optional) Initial value for the lake surface water temperature.
        """
        # calculate the proportion of currrent day of year to the total number of days on the year
        doy = np.array([dt.dayofyear for dt in period])
        tdoy = np.array([pd.Timestamp(dt.year, 12, 31).dayofyear for dt in period])
        t_ty = doy/tdoy
        
        # turn ta to numpy array
        if isinstance(ta, pd.Series):
            ta = ta.to_numpy().ravel()
        elif isinstance(ta, np.array):
            ta = ta.ravel()
        elif isinstance(ta, list):
            ta = np.array(ta).ravel()
        else:
            raise ValueError("Air temp should be one of the numpy array, pandas series and python list")
        
        # use model params
        params = self.params[0]
        
        if self._model_version == "6p":
            run_air2water = run_air2water_6p
        elif self._model_version == "8p":
            run_air2water = run_air2water_8p
        else:
            raise ValueError("Select correct model version: '6p' or '8p'")
        
        # call simulation function given the parameter
        tw = run_air2water(ta, t_ty, th, tw_init, tw_ice, params)

        # build pandas dataframe
        tw = pd.DataFrame(tw, index = period, columns = ["tw_sim"])
        
        return tw

    def pso_fit(self,
            tw_obs,
            ta,
            period,
            obj_func = "MSE",
            th = 4.0,
            tw_init = 1.0,
            tw_ice = 0.0,
            swarm_size = 100,
            n_cpus = 1,
            iteration_num = 500,
            options = {'c1': 0.5, 'c2': 0.3, 'w': 0.9},
            ):
        """This function uses Particle Swarm Optimization algorithm to search for the best parameter combination.

        Args:
            tw_obs (np.array): A numpy array of lake surface water temperature observations. Missing value should be set as np.NaN
            ta (np.array): A numpy array of air temperature. Should be continous.
            period (pandas.Series): simulation period, use pd.Timestamp to indicate time. Preferably pandas series.
            obj_func (str, optional): Objective function for calibration. Could be `MSE`, `NSE`. Defaults to `MSE`.
            th (float, optional): _description_. Defaults to 4.0.
            tw_init (float, optional): _description_. Defaults to 1.0.
            tw_ice (float, optional): _description_. Defaults to 0.0.
            swarm_size (int, optional): _description_. Defaults to 100.
            n_cpus (int, optional): _description_. Defaults to 1.
            iteration_num (int, optional): _description_. Defaults to 500.
            options (dict, optional): _description_. Defaults to {'c1': 0.5, 'c2': 0.3, 'w': 0.9}.

        Returns:
            loss and parameters
        """
        
        # calculate the proportion of currrent day of year to the total number of days on the year
        doy = np.array([dt.dayofyear for dt in period])
        tdoy = np.array([pd.Timestamp(dt.year, 12, 31).dayofyear for dt in period])
        t_ty = doy/tdoy
        
        # turn ta to numpy array
        if isinstance(ta, pd.Series):
            ta = ta.to_numpy().ravel()
        elif isinstance(ta, np.array):
            ta = ta.ravel()
        elif isinstance(ta, list):
            ta = np.array(ta).ravel()
        else:
            raise ValueError("Air temp should be one of the numpy array, pandas series and python list")
        
        # Cast initial state as float
        tw_init = float(tw_init)
        # pack input arguments for scipy optimizer
        input_args = (ta, t_ty, th, tw_init, tw_ice, tw_obs, self._dtype, self._model_version, n_cpus, obj_func)
        constraints = (np.array([self._default_bounds[p][0] for p in self._param_list]),
                       np.array([self._default_bounds[p][1] for p in self._param_list]))
        
        # initialize a optimizer
        optimizer = GlobalBestPSO(n_particles=swarm_size, 
                                  dimensions=len(self._param_list), 
                                  options=options, 
                                  bounds=constraints, 
                                  )
        
        cost, joint_vars = optimizer.optimize(_loss_pso, iters = iteration_num, input_args = input_args)
        
        return cost, joint_vars
        
def _loss_pso(X, input_args):
    """Return the loss value for the current parameter set.
    The shape of X is (n_particles)    
    """
    # Unpack static arrays
    ta = input_args[0]
    t_ty = input_args[1]
    th = input_args[2]
    tw_init = input_args[3]
    tw_ice = input_args[4]
    tw_obs = input_args[5]
    dtype = input_args[6]
    model_version = input_args[7]
    n_cpus = input_args[8]
    obj_func = input_args[9]
    
    # Create a custom numpy array of the model parameters
    # number of particles
    n_particles = X.shape[0]    
    params = np.zeros(n_particles, dtype=dtype)    
    if model_version == "6p":
        params['a1'] = X[:, 0]
        params['a2'] = X[:, 1]
        params['a3'] = X[:, 2]
        params['a4'] = X[:, 3]
        params['a5'] = X[:, 4]
        params['a6'] = X[:, 5]
        run_air2water = run_air2water_6p
    elif model_version == "8p":
        params['a1'] = X[:, 0]
        params['a2'] = X[:, 1]
        params['a3'] = X[:, 2]
        params['a4'] = X[:, 3]
        params['a5'] = X[:, 4]
        params['a6'] = X[:, 5]
        params['a7'] = X[:, 6]
        params['a8'] = X[:, 7]
        run_air2water = run_air2water_8p
    else:
        raise ValueError("Select correct model version, '6p' or '8p'")
    
    # run air2water model for particle times get the results
    if n_cpus > 1:
        # parallel
        foo_ = partial(run_air2water, ta, t_ty, th, tw_init, tw_ice,)
        tws = Parallel(n_jobs=n_cpus)(delayed(foo_)(i) for i in params)
    elif n_cpus == 1:
        tws = [run_air2water(ta, t_ty, th, tw_init, tw_ice, params[n]) for n in range(n_particles)]
    else:
        raise ValueError("Choose a positive number of the n_cpus")
    
    if obj_func == "MSE":    
        # Calculate the simulated lake surface water temperature and calculate the mse
        loss_values = np.array([calc_mse(tw_obs, tw) for tw in tws])
    elif obj_func == "NSE":
        # Note the NSE here will be -NSE
        loss_values = np.array([-calc_nse(tw_obs, tw) for tw in tws])
    elif obj_func == "R2":
        loss_values = np.array([-calc_r2(tw_obs, tw) for tw in tws])
    else:
        raise ValueError(f"Current objective functions do not support {obj_func}. Consider use MSE, NSE or R2")
    # # replace the nan value with 999
    # loss_values = np.nan_to_num(loss_values, nan = 999, posinf=1000)
    # print(loss_values)
    return loss_values